/* io/console.c */

#include "externs.h"


/*
 *
 * Reserved for Linux Console Color Device (screen fading via spells/timeofday)

static unsigned char default_rgb[48]={
     0, 0, 0,     170, 0, 0,    0, 170, 0,    170, 85, 0,
     0, 0, 170,   170, 0, 170,  0, 170, 170,  170, 170, 170, 

     85, 85, 85,  255, 85, 85,  85, 255, 85,  255, 255, 85,
     85, 85, 255, 255, 85, 255, 85, 255, 255, 255, 255, 255};

static char tones[16]={0, 4, 6, 2, 3, 5, 7, 8, 1, 10, 12, 14, 9, 11, 13, 15};
 */

static char fadetab[40]={33, 27, 12, 9, 15, 19, 0, 31, 11, 16,
                         22, 34, 14, 4, 29, 17, 37, 20, 10, 38,
                         21, 25, 36, 3, 35, 30, 6, 13, 8, 39,
                         26, 18, 1, 23, 5, 32, 2, 28, 7, 24};

static char highfont[128]=
  "                                "
  "        ?C#  !<>   ||          ."
  "`^.|-+                |~_'. ||||"
  "                =     *|^.>*  - ";

/* gets #cols from session <d>, or player <player> if not connected */
int get_cols(DESC *d, dbref player)
{
  player=db[player].owner;

  if(!d)
    d=Desc(player);
  if(!d || d->cols < 4)
    return db[player].data->cols;
  return d->cols;
}

/* gets #rows from session <d>, or player <player> if not connected */
int get_rows(DESC *d, dbref player)
{
  player=db[player].owner;

  if(!d)
    d=Desc(player);
  if(!d || d->rows < 4)
    return db[player].data->rows;
  return d->rows;
}

/* console reset */
void clear_screen(DESC *d, int sbar)
{
  char buf[512];
  int a, b;

  if(Invalid(d->player)) {
    output2(d, "\ec\e(U\e[1;5]\e[2J\e[H\n");
    return;
  }

  /* clear screen without ansi */
  if(!Invalid(d->player) && !terminfo(d->player, TERM_ANSI)) {
    b=get_rows(d, d->player)-1;
    for(a=0;a<b;a++)
      buf[a]='\n';
    buf[a]='\0';
    output2(d, buf);
    return;
  }

  /* perform ansi reset escape sequences */
  output2(d, tprintf("\e[m\e[r%s\e[1;5]\e[2J\e[%dH\n", (terminfo(d->player,
          TERM_HIGHFONT) || terminfo(d->player, TERM_GRAPHICS))?"\e(U":"",
          sbar?4:1));
}

/* fade away screen with 2x1 character blocks */
void fade_screen(DESC *d)
{
  char buf[256], *s=buf;
  int i, x, y, count=0, cols, rows;

  if(!terminfo(d->player, TERM_ANSI))
    return;
  cols=get_cols(d, d->player);
  rows=get_rows(d, d->player);

  for(i=0;i<40;i++)
    for(y=1;y<rows;y++)
      for(x=(fadetab[(fadetab[i]+y)%40]*2)+1; x<cols; x+=80) {
        s+=sprintf(s, "\e[%d;%dH  ", y, x);
        if(++count == 11) {
          output2(d, tprintf("\e7%s\e8\e[A\n", buf));
          s=buf, count=0;
        }
      }

  if(s > buf)
    output2(d, tprintf("\e7%s\e8\e[A\n", buf));
}


/* Terminal Configuration Section */

struct term_config term_config[]={
  {"Rows",	0},
  {"Columns",	0},
  {"Ansi",	TERM_ANSI},
  {"Highfont",	TERM_HIGHFONT},
  {"Graphics",	TERM_GRAPHICS},
  {"Wordwrap",	TERM_WORDWRAP},
  {"Beeps",	TERM_BEEPS},
  {"Sound",	TERM_SOUND},
  {"Icons",	TERM_ICONS},
  {"Reginfo",	TERM_REGINFO},
  {0}};


/* terminal configuration */
void do_term(dbref player, char *arg1, char *arg2)
{
  DESC *d;
  dbref thing=db[player].owner;
  int a=0, b;

  /* Match configuration variable */
  if(*arg1) {
    for(a=0;term_config[a].name;a++)
      if(string_prefix(term_config[a].name, arg1))
        break;
    if(!term_config[a].name) {
      if((thing=lookup_player(arg1)) <= NOTHING) {
        notify(player, "Unknown terminal setting.");
        return;
      } else if(!controls(player, thing, POW_EXAMINE)) {
        notify(player, "Permission denied.");
        return;
      }
    }
  }

  d=Desc(thing);

  /* Display terminal configuration */
  if(!*arg1 || !term_config[a].name) {
    char buf[16];

    notify(player, "\e[1;32m%s%s Terminal Configuration:\e[m",
           db[thing].name, possname(db[thing].name));
    for(a=0;term_config[a].name;a++) {
      if(a < 2)
        sprintf(buf, "%d", a?get_cols(d, thing):get_rows(d, thing));
      else
        strcpy(buf, (db[thing].data->term & term_config[a].flag)?"On":"Off");

      notify(player, "\e[1;36m%12.12s: \e[37m%s\e[m", term_config[a].name,buf);
    }
    return;
  }

  /* Determine correct argument */
  if(a < 2) {
    b=atoi(arg2);
    if(b < 4 || b > 250) {
      notify(player, "%s may only be between 4 and 250.", term_config[a].name);
      return;
    }
  } else {
    if(!strcasecmp(arg2, "Off"))
      b=0;
    else if(!strcasecmp(arg2, "On"))
      b=1;
    else {
      notify(player, "%s may either be On or Off.", term_config[a].name);
      return;
    }
  }

  /* do additional things if removing ANSI */
  if(a == 2 && !b)
    db[thing].data->term &= ~TERM_SOUND;

  /* Turning off Highfont turns off Graphics too */
  if(a == 3 && !b)
    db[thing].data->term &= ~TERM_GRAPHICS;

  if(!a) {  // Resize Rows
    db[thing].data->rows=b;
    if(d)
      d->rows=b;
  } else if(a == 1) {  // Resize Columns
    db[thing].data->cols=b;
    if(d) {
      /* Special: Resize screen if #cols = 80 or 132 */
      if((b == 80 || b == 132) && terminfo(thing, TERM_ANSI))
        output2(d, tprintf("\e[?3%c\e[A\n", b<132?'l':'h'));

      d->cols=b;
    }
  } else {  // All other +term variables
    if(b) {
      db[thing].data->term |= term_config[a].flag;
      if(a == 4)
        db[thing].data->term |= term_config[a-1].flag;
    } else
      db[thing].data->term &= ~term_config[a].flag;
  }

  if(a < 2)
    notify(player, "%s set to %d.", term_config[a].name, b);
  else
    notify(player, "%s turned %s.",term_config[a].name, b?"On":"Off");
}

/* refixes player's screen */
void do_cls(dbref player)
{
  DESC *d;

  if(Typeof(player) == TYPE_PLAYER && (d=Desc(player)))
    clear_screen(d, 0);
}


/* player console settings */
void do_console(dbref player, char *name, char *options)
{
  static struct {
    char *name;
    int bitmask;	// Connection flag bitmask
    int inverse;	// Set flag when the option is disabled
  } cflags[]={
    {"Input",	C_INPUTOFF,	1},
    {"Output",	C_OUTPUTOFF,	1},
    {"Sbar",	C_SBAROFF,	1},
    {"Reconnect"},
    {"Fade"},
    {0}};

  DESC *d;
  char *r, *s=options;
  int a, inv, thing;

  /* match login descriptor to player */
  if((thing=match_thing(player, name, MATCH_EVERYTHING)) == NOTHING)
    return;
  if(Typeof(thing) != TYPE_PLAYER) {
    notify(player, "Console settings can only be changed for players.");
    return;
  } if(!controls(player, thing, POW_CONSOLE)) {
    notify(player, "Permission denied.");
    return;
  } if(thing == GOD) {
    notify(player, "Cannot change console options for %s.",
           unparse_object(player, thing));
    return;
  } if(!(d=Desc(thing)) || !(d->flags & C_CONNECTED)) {
    notify(player, "That player is not connected.");
    return;
  }

  while((r=parse_up(&s,' '))) {
    if(*r == '!') {
      inv=1;
      r++;
    } else
      inv=0;
    if(!*r)
      continue;

    for(a=0;cflags[a].name;a++)
      if(string_prefix(cflags[a].name, r))
        break;
    if(!cflags[a].name) {
      notify(player, "No such console option '%s'.", r);
      continue;
    }
    
    /* screen animation triggering */
    if(!cflags[a].bitmask) {
      if(inv) {
        notify(player, "Console commands cannot be inverted.");
        continue;
      }

      /* trigger the animation */
      switch(a) {
        case 3:
          process_output(d);
          output2(d, "*** Reconnecting ***\n\n");
          d->flags |= C_RECONNECT|C_LOGOFF;
          break;
        case 4:
          fade_screen(d);
          break;
      }

      if(player != thing && !Quiet(player))
        notify(player, "Command triggered.");
      continue;
    }

    /* descriptor flag setting */
    if((!(inv ^ cflags[a].inverse) && (d->flags & cflags[a].bitmask)) ||
         ((inv ^ cflags[a].inverse) && !(d->flags & cflags[a].bitmask))) {
      if(!Quiet(player))
        notify(player, "Option already %sset.", inv?"un":"");
      continue;
    }

    if(inv ^ cflags[a].inverse)
      d->flags &= ~cflags[a].bitmask;
    else
      d->flags |= cflags[a].bitmask;

    if(!Quiet(player))
      notify(player, "Option %sset.", inv?"un":"");
  }
}


/* Text Emulation & Translation */

char *term_emulation(DESC *d, char *msg, int *len)
{
  static unsigned char buf[16384];
  unsigned char *s=buf;
  int terminfo;
  
  terminfo=(game_online && !Invalid(d->player))?db[d->player].data->term:
           (TERM_BEEPS|TERM_ANSI|TERM_HIGHFONT);

  /* apply wordwrap to output string, except for lines with special codes */
  if((terminfo & TERM_WORDWRAP) && (msg[0] != '\e' || msg[1] == '['))
    msg=wordwrap(d->player, msg);

  /* strip ansi escape codes */
  if(!(terminfo & TERM_ANSI))
    msg=ansi_strip(msg);

  /* format message for output */
  for(;*msg;msg++) {
    /* strip beeps */
    if(*msg == '\a' && !(terminfo & TERM_BEEPS))
      continue;

    /* transpose LF into CRLF */
    if(*msg == '\n')
      *s++='\r';
    *s++=*msg;
  }
  *s='\0';
  *len=s-buf;

  /* change 8-bit font to 7-bit for players */
  if(!(terminfo & TERM_HIGHFONT))
    for(s=buf;*s;s++)
      if(*s > 127) {
        if(*s == 255) {  /* Skip Telnet IAC */
          if(*(s+1)) s++;
          if(*(s+1)) s++;
        } else
          *s=highfont[(int)*s - 128];
      }

  return buf;
}

// Silently strips ansi escape sequences from string.
//
char *ansi_strip(char *str)
{
  static char buff[8192];
  char *s, *t=buff;

  for(s=str;*s;s++) {
    if(*s != '\e') {  /* copy everything that isn't an escape character */
      *t++=*s;
      continue;
    }

    /* we found an escape character */
    if(!*(++s))
      break;

    /* 1-digit escape sequence */
    if(*s != '[' && *s > 47)
      continue;
    s++;

    /* 2-digit sequence */
    if(*s < 48) {
      if(!*(++s))
        break;
      continue;
    }

    /* variable length escape code; terminates with a character > 63 */
    while(*s && *s < 64)
      s++;
    if(!*s)
      break;
  }

  *t='\0';
  return buff;
}

// Enable wordwrap around ansi characters.
//
char *wordwrap(dbref player, char *str)
{
  static char buf[8192];
  char *r=buf, *p=NULL, *s=str;
  int wlen=0, llen=0, width=get_cols(NULL, player)-1;

  while(*s && r-buf < 8000) {
    if(*s == '\n') {	// Newline
      *r++=*s++;
      llen=wlen=0;
      p=NULL;
      continue;
    }

    if(*s == '\t') {    // Pseudo-tab
      if(llen < width)
        llen+=(llen%8)?8-(llen%8):8;
      if(llen >= width) {
        *r++='\n';
        s++;
        llen=wlen=0;
        p=NULL;
      } else
        *r++=*s++;
      continue;
    }

    if(*s == ' ') {	// Space
      if(llen == width) {
        *r++='\n';
        if(!wlen)
          *r++=*s;
	s++;
        llen=wlen=0;
        p=NULL;
      } else {
        p=r;
        *r++=*s++;
        llen++;
        wlen=0;
      }
      continue;
    }

    if(*s == '\e') {	// Escape character; copy ansi codes
      if(!*(s+1))
        break;
      *r++=*s++;
      if(*s != '[' && *s > 47) {
        *r++=*s++;
        continue;
      }
      *r++=*s++;
      if(!*s)
        break;
      if(*s < 48) {
        *r++=*s++;
        continue;
      }
      while(*s && r-buf < 8000 && *s < 64)
        *r++=*s++;
      if(!*s)
        break;
      *r++=*s++;
      continue;
    }

    if(llen == width) {
      if(p) {
        *p='\n';
        llen=wlen;
        p=NULL;
      } else {
        *r++='\n';
        p=NULL;
        llen=wlen=0;
      }
    }

    *r++=*s++;
    llen++;
    wlen++;
  }
  *r='\0';
  return buf;
}


/* Color Manipulation */

int default_color=7;

// Return an integer color code based on a list of specified flags or numbers.
//
int color_code(int prev, char *s)
{
  static char flags[33]="h+f#u-i!*/InxrgybmvpcwNXRGYBMVPCW";
  static int attr[33]={8, 8, 128, 128, 256, 256, 512, 512, 1024, 2048, 4096,
                       -1, 0, 1,  2,  3,  4,  5,  5,  5,  6,  7,
                       -1, 0, 16, 32, 48, 64, 80, 80, 80, 96, 112};
  int a, color=prev;

  /* treat numbers as a base starting color */
  if(isdigit(*s)) {
    if((a=atoi(s)))
      color=a;
    while(isdigit(*++s));  // Skip pointer over digits to start of flags
  }

  /* parse flags string */
  for(;*s;s++)
    for(a=0;a<33;a++)
      if(flags[a] == *s) {
        switch(a) {
          case 11:          color = prev; break;     // Default @color.
	  case 22:          color = 7; break;        // All attributes off.
          case 12 ... 21:   color = (color & ~7) + attr[a]; break;
          case 23 ... 32:   color = (color & ~112) + attr[a]; break;
          default:          color |= attr[a];
        }
        break;
      }

  return color;
}

// Returns a color string containing flags used to compose color attributes.
//
char *unparse_color_code(int color)
{
  static char flags[13]="RGYBMCW#-!*/I";
  static char result[16];
  char *s=result;
  int i;

  s+=sprintf(s, "%d", color & 15);
  if(color & 112)
    *s++=flags[((color & 112) >> 4)-1];

  for(i=7;i<13;i++)
    if(color & (1 << i))
      *s++=flags[i];

  *s='\0';
  return result;
}

// Outputs an ANSI escape code to change the text color from <prev> to <color>.
// Can be called up to 4 times within one sprintf() function.
//
char *textcolor(int prev, int color)
{
  static char buf[4][24];
  static int cycle;
  char *result, *s;
  int set;

  if(!((prev^color) & (4096|2048|512|256|255)))  // No change?
    return "";
  if((color & ~1024) == 7)  // Default attributes
    return "\e[m";

  result=s=buf[cycle];
  cycle=(cycle+1) & 3;

  *s++='\e', *s++='[';
  /* Check if we're unsetting any attributes */
  if(prev & (~color & (4096|2048|512|256|128|8))) {
    *s++='0', *s++=';';
    prev=7;
  }

  set=color & ~prev;
  if(set & 4096) *s++='8', *s++=';';  // Invisible
  if(set & 2048) *s++='2', *s++=';';  // Half-intensity
  if(set &  512) *s++='7', *s++=';';  // Inverse
  if(set &  256) *s++='4', *s++=';';  // Underline/Italics
  if(set &  128) *s++='5', *s++=';';  // Blinking
  if(set &    8) *s++='1', *s++=';';  // Hilighted

  if((prev^color) & 7)
    *s++='3', *s++='0'|(color & 7), *s++=';';  // Foreground color changed
  if(((prev^color) >> 4) & 7)
    *s++='4', *s++='0'|((color>>4) & 7), *s++=';';  // Background color

  *(s-1)='m', *s='\0';  // The last semicolon gets replaced with an 'm'
  return result;
}

int prev_rainbow;  /* used to get the last color of the rainbow text */

char *rainbowmsg(char *s, int color, int size)
{
  static int seq[16]={4, 3, 6, 11, 5, 13, 14, 6, 6, 1, 2, 10, 4, 9, 12, 11};
  static char temp[8192];
  int prev=default_color;
  char *t=temp;

  /* If the first color is Normal, set it to Light Violet */
  if((color & 15) == 7)
    color=(color & ~15)|13;

  while(*s && t-temp < size-42) {
    strcpy(t, textcolor(prev, color));
    t+=strlen(t);
    *t++=*s++;
    if(*s != ' ') {
      prev=color;
      color=(color & ~15)|seq[color & 15];
    }
  }
  prev_rainbow=prev;
  strcpy(t, textcolor(prev, default_color));
  return temp;
}
