/* prog/eval.c - Function Parser */

#include "externs.h"
#include <math.h>

static int func_recur;  // Current depth level in nested functions.
static int func_total;  // Number of total functions processed in command.

#define FUNCTION(x) static void fun_##x(char *buff, char **args, \
                                        dbref privs, dbref doer, int nargs)

#define INIT_MATCH(x,y)    dbref x=match_thing(privs, y, MATCH_EVERYTHING)
#define INIT_CMATCH(x,y,z) dbref x=match_controlled(privs, y, z)

#define ISTRUE(x) (*(x) && strncmp(x, "#-", 2) && (!isdigit(*(x)) || atoi(x)))

#define ADDWORD(buff, s, format, args...) \
 ({  if(s != buff) { \
       if(s-buff > 7980) { strcpy(s, " ..."); return; } \
       *s++=' '; \
     } s += sprintf(s, format, ## args);  })

#define DELIMCH(x,y,ch)		char x=((nargs > y && *args[y] > 32 && \
                                         *args[y] < 127) ? *args[y] : ch)
#define DELIM(x,y) DELIMCH(x,y,' ')

#define ERR_PERM	"#-1 Permission denied."
#define ERR_PLAYER	"#-1 No such player."
#define ERR_RANGE	"#-1 Out of range."
#define ERR_LOC		"#-1 Too far away to see."
#define ERR_CONN	"#-1 Not connected."
#define ERR_DATE	"#-1 Invalid date."
#define ERR_ENV		"#-1 Invalid global register."


/* Set 1: Basic Math *********************************************************/

FUNCTION(add) {
  int a, result=0;
  char *r, *s;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      result+=atoi(r);
  }
  sprintf(buff, "%d", result);
}

FUNCTION(fadd) {
  int a;
  char *r, *s;
  double result=0;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      result+=atof(r);
  }
  print_float(buff, result);
}

FUNCTION(mul) {
  int a, result=1;
  char *r, *s;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      if(*r)
        result*=atoi(r);
  }
  sprintf(buff, "%d", result);
}

FUNCTION(fmul) {
  int a;
  char *r, *s;
  double result=1;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      if(*r)
        result*=atof(r);
  }
  print_float(buff, result);
}

FUNCTION(sub) { sprintf(buff, "%d", atoi(args[0])-atoi(args[1])); }
FUNCTION(fsub) { print_float(buff, atof(args[0])-atof(args[1])); }
FUNCTION(div) { if(!atoi(args[1])) strcpy(buff, "inf");
                  else sprintf(buff, "%d", atoi(args[0])/atoi(args[1])); }
FUNCTION(fdiv) { if(!atof(args[1])) strcpy(buff, "inf");
                   else print_float(buff, atof(args[0])/atof(args[1])); }
FUNCTION(mod) { if(!atoi(args[1])) strcpy(buff, "0");
                  else sprintf(buff, "%d", atoi(args[0])%atoi(args[1])); }

FUNCTION(min) {
  int a, count=0;
  char *r, *s;
  double result=0;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      if(*r && (!count++ || atof(r) < result))
        result=atof(r);
  }
  print_float(buff, result);
}

FUNCTION(max) {
  int a, count=0;
  char *r, *s;
  double result=0;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s,' ')))
      if(*r && (!count++ || atof(r) > result))
        result=atof(r);
  }
  print_float(buff, result);
}

FUNCTION(avg) {
  int a, count=0;
  char *r, *s;
  double result=0;

  for(a=0;a<nargs;a++) {
    s=args[a];
    while((r=parse_up(&s, ' ')))
      if(*r) {
        result+=atof(r);
        count++;
      }
  }

  print_float(buff, count?(result/count):0);
}

/* Set 2: Predicate Calculus / Logic *****************************************/

FUNCTION(gt)   { sprintf(buff, "%d", atof(args[0]) >  atof(args[1])); }
FUNCTION(gteq) { sprintf(buff, "%d", atof(args[0]) >= atof(args[1])); }
FUNCTION(lt)   { sprintf(buff, "%d", atof(args[0]) <  atof(args[1])); }
FUNCTION(lteq) { sprintf(buff, "%d", atof(args[0]) <= atof(args[1])); }
FUNCTION(eq)   { sprintf(buff, "%d", atof(args[0]) == atof(args[1])); }
FUNCTION(neq)  { sprintf(buff, "%d", atof(args[0]) != atof(args[1])); }

FUNCTION(truth) { sprintf(buff, "%d", ISTRUE(args[0])); }
FUNCTION(not) { sprintf(buff, "%d", !ISTRUE(args[0])); }
FUNCTION(if) { strcpy(buff, ISTRUE(args[0])?args[1]:""); }
FUNCTION(ifelse) { strcpy(buff, ISTRUE(args[0])?args[1]:args[2]); }
FUNCTION(iftrue) { strcpy(buff, ISTRUE(args[0])?args[0]:args[1]); }
FUNCTION(between) { sprintf(buff, "%d", (atof(args[0])>=atof(args[1]) &&
                                         atof(args[0])<=atof(args[2]))); }

FUNCTION(band) { sprintf(buff, "%d", atoi(args[0]) & atoi(args[1])); }
FUNCTION(bnand) { sprintf(buff, "%d", atoi(args[0]) & ~atoi(args[1])); }
FUNCTION(bor) { sprintf(buff, "%d", atoi(args[0]) | atoi(args[1])); }
FUNCTION(bxor) { sprintf(buff, "%d", atoi(args[0]) ^ atoi(args[1])); }
FUNCTION(land) { sprintf(buff, "%d", ISTRUE(args[0]) && ISTRUE(args[1])); }
FUNCTION(lor) { sprintf(buff, "%d", ISTRUE(args[0]) || ISTRUE(args[1])); }
FUNCTION(lxor) { sprintf(buff, "%d", ISTRUE(args[0]) ^ ISTRUE(args[1])); }

FUNCTION(shl) { sprintf(buff, "%d", atoi(args[0]) << atoi(args[1])); }
FUNCTION(shr) { sprintf(buff, "%d", atoi(args[0]) >> atoi(args[1])); }
FUNCTION(ffs) { sprintf(buff, "%d", ffs(atoi(args[0]))); }

FUNCTION(bitcount) { sprintf(buff, "%d", BITCOUNT(atoi(args[0]))); }

/* Set 3: Math Library *******************************************************/

FUNCTION(e) { strcpy(buff, "2.718281828"); }
FUNCTION(pi) { strcpy(buff, "3.141592654"); }
FUNCTION(sin) { print_float(buff, sin(atof(args[0]))); }
FUNCTION(cos) { print_float(buff, cos(atof(args[0]))); }
FUNCTION(tan) { print_float(buff, tan(atof(args[0]))); }
FUNCTION(asin) { if(fabs(atof(args[0])) > 1) strcpy(buff, "inf");
                   else print_float(buff, asin(atof(args[0]))); }
FUNCTION(acos) { if(fabs(atof(args[0])) > 1) strcpy(buff, "inf");
                   else print_float(buff, acos(atof(args[0]))); }
FUNCTION(atan) { print_float(buff, atan(atof(args[0]))); }

FUNCTION(exp) { print_float(buff, exp(atof(args[0]))); }
FUNCTION(pow) { print_float(buff, pow(atof(args[0]),atof(args[1]))); }
FUNCTION(sqrt) { if(atof(args[0]) < 0) strcpy(buff, "nan");
                   else print_float(buff, sqrt(atof(args[0]))); }
FUNCTION(ln) { if(atof(args[0]) <= 0) strcpy(buff, "nan");
                 else print_float(buff, log(atof(args[0]))); }

FUNCTION(neg) { print_float(buff, atof(args[0])?-atof(args[0]):0); }
FUNCTION(abs) { print_float(buff, fabs(atof(args[0]))); }
FUNCTION(sign) { sprintf(buff, "%d", atof(args[0])>0?1:atof(args[0])<0?-1:0); }

FUNCTION(int) { sprintf(buff, "%d", atoi(args[0])); }
FUNCTION(ceil) { sprintf(buff, "%.0f", ceil(atof(args[0]))); }
FUNCTION(floor) { sprintf(buff, "%.0f", floor(atof(args[0]))); }

FUNCTION(degrees) { print_float(buff, degrees(atof(args[0]),atof(args[1]))); }
FUNCTION(dist2d) { print_float(buff, sqrt(pow(atof(args[0])-atof(args[2]),2)+
                        pow(atof(args[1])-atof(args[3]),2))); }
FUNCTION(dist3d) { print_float(buff, sqrt(pow(atof(args[0])-atof(args[3]),2)+
                        pow(atof(args[1])-atof(args[4]),2)+
                        pow(atof(args[2])-atof(args[5]),2))); }

FUNCTION(round) {
  int range=(nargs==2)?atoi(args[1]):0;
  sprintf(buff, "%.*f", range<0?0:range>9?9:range, atof(args[0]));
}

FUNCTION(log) {
  double base=(nargs==2)?atof(args[1]):10.0;

  if(atof(args[0]) <= 0 || base <= 0)
    strcpy(buff, "nan");
  else
    print_float(buff, log(atof(args[0]))/log(base));
}

FUNCTION(factor) {
  int count=0, d=3, q, n=atoi(args[0]);
  char *s=buff;

  if(n >= -1 && n <= 1) {
    sprintf(buff, "%d", n);
    return;
  } if(n < 0) {
    *s++='-';
    n=-n;
  }

  /* knock off all 2's */
  while(!(n & 1) && n) {
    if(count++)
      *s++=' ';
    *s++='2';
    n >>= 1;
  }
  *s='\0';

  /* find primes starting at 3 and increment by 2 */
  do {
    while(n == (q=n/d) * d) {
      sprintf(s, "%s%d", count++?" ":"", d);
      s+=strlen(s);
      n = q;
    }
  } while((d += 2) <= q);

  /* a prime number remains */
  if(n > 1)
    sprintf(s, "%s%d", count++?" ":"", n);
}

/* Set 4: Game Runtime Configuration *****************************************/

FUNCTION(mudname)  { strcpy(buff, MUD_NAME); }
FUNCTION(version)  { strcpy(buff, mud_version); }
FUNCTION(config)   { get_config(privs, buff, args[0]); }
FUNCTION(uptime)   { sprintf(buff, "%ld", now-mud_up_time); }
FUNCTION(restarts) { sprintf(buff, "%d", nstat[NS_EXEC]-1); }
FUNCTION(dbtop)    { sprintf(buff, "#%d", db_top-1); }
FUNCTION(upfront)  { sprintf(buff, "#%d", get_upfront()); }
FUNCTION(cpu)      { sprintf(buff, "%.3f", (double)cpu_percent/1000); }
FUNCTION(cputime)  { sprintf(buff, "%.2f", (double)get_cputime()/1000); }

FUNCTION(netstat) {
  const char *ns[NET_STATS]={
    "FILES", "OUTPUT", "INPUT", "EXEC", "CRASHES", "DBSAVE", "LOGINS",
    "CONNECTS", "FAILED", "NEWCR", "GUESTS", "NCMDS", "NPCMDS", "NQCMDS",
    "NFCMDS", "QUEUE", "OVERRUNS", "MAIL", "EMAIL", "ERECV", "QUERY", "RESOLV",
  };

  int a;

  for(a=0;a<NET_STATS && ns[a];a++)
    if(!strcasecmp(ns[a], args[0]))
      break;
  if(a == NET_STATS)
    strcpy(buff, "#-1 No such variable.");
  else
    sprintf(buff, "%d", nstat[a]);
}

/* Set 5: Global Weather Almanac *********************************************/

FUNCTION(wtime) { strcpy(buff, wtime(nargs?atol(args[0]):-1L,0)); }
FUNCTION(wdate) { strcpy(buff, wtime(nargs?atol(args[0]):-1L,1)); }
FUNCTION(timeofday) { sprintf(buff, "%d", TIMEOFDAY); }

/* Set 6: Time and Session Management ****************************************/

FUNCTION(time) {
  long tt=nargs?get_date(args[0], privs, 0L):now;
  strcpy(buff, (tt == -1L)?ERR_DATE:mktm(privs, tt));
}

FUNCTION(date) {
  long tt=nargs?get_date(args[0], privs, 0L):now;
  strcpy(buff, (tt == -1L)?ERR_DATE:mkdate(mktm(privs, tt)));
}

FUNCTION(rtime) {
  long tt=nargs?get_date(args[0], privs, 0L):now;
  if(tt == -1L)
    strcpy(buff, ERR_DATE);
  else {
    char *s=mktm(privs, tt);
    sprintf(buff, "%02d%-6.6s %cM", (atoi(s+11)+11)%12+1, s+13,
            (atoi(s+11) > 11)?'P':'A');
  }
}

FUNCTION(xtime) {
  if(!nargs) {
    struct timeval current;
    gettimeofday(&current, NULL);
    sprintf(buff, "%f", (double)current.tv_sec+
            ((double)current.tv_usec/1000000));
  } else {
    long tt=get_date(args[0], privs, 0L);
    if(tt == -1L)
      strcpy(buff, ERR_DATE);
    else
      sprintf(buff, "%ld", tt);
  }
}

FUNCTION(laston) {
  dbref thing=lookup_player(args[0]);

  if(thing == AMBIGUOUS)
    thing=privs;
  if(thing == NOTHING || Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else
    sprintf(buff, "%ld", db[thing].data->last);
}

FUNCTION(lastoff) {
  dbref thing=lookup_player(args[0]);

  if(thing == AMBIGUOUS)
    thing=privs;
  if(thing == NOTHING || Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else if(!controls(privs, thing, POW_WHO))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%ld", db[thing].data->lastoff);
}

FUNCTION(tml) { strcpy(buff, time_format(TML, atol(args[0]))); }
FUNCTION(tms) { strcpy(buff, time_format(TMS, atol(args[0]))); }
FUNCTION(tmf) { strcpy(buff, time_format(TMF, atol(args[0]))); }
FUNCTION(tma) { strcpy(buff, time_format(TMA, atol(args[0]))); }

static void io_list(char *buff, dbref privs, char *name, int type)
{
  DESC *d;
  dbref who;

  if((who=lookup_player(name)) == AMBIGUOUS)
    who=privs;
  if(who == NOTHING || Typeof(who) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else if(!controls(privs, who, POW_WHO) && (type > 2 ||
          !could_doit(privs, who, A_HLOCK)))
    strcpy(buff, ERR_PERM);
  else if(!(d=Desc(who)))
    strcpy(buff, ERR_CONN);
  else switch(type) {
    case 1: sprintf(buff, "%ld", now - d->last_time); return;
    case 2: sprintf(buff, "%ld", now - d->login_time); return;
    case 3: strcpy(buff, d->addr); return;
    case 4: sprintf(buff, "%d", d->concid); return;
    case 5: sprintf(buff, "%d", d->port); return;
    case 6: sprintf(buff, "%d", d->cmds); return;
    case 7: sprintf(buff, "%d", d->input); return;
    case 8: sprintf(buff, "%d", d->output); return;
    default: *buff='\0';
  }
}

FUNCTION(idle)   { io_list(buff, privs, args[0], 1); }
FUNCTION(onfor)  { io_list(buff, privs, args[0], 2); }
FUNCTION(host)   { io_list(buff, privs, args[0], 3); }
FUNCTION(concid) { io_list(buff, privs, args[0], 4); }
FUNCTION(port)   { io_list(buff, privs, args[0], 5); }
FUNCTION(cmds)   { io_list(buff, privs, args[0], 6); }
FUNCTION(input)  { io_list(buff, privs, args[0], 7); }
FUNCTION(output) { io_list(buff, privs, args[0], 8); }

FUNCTION(lwho) {
  DESC *d;
  char *s=buff;
 
  *buff=0;
  for(d=descriptor_list;d;d=d->next)
    if((d->flags & C_CONNECTED) && (controls(privs, d->player, POW_WHO) ||
       could_doit(privs, d->player, A_HLOCK)))
      ADDWORD(buff, s, "#%d", d->player);
}

/* Set 7: Database Values ****************************************************/

FUNCTION(num) {
  sprintf(buff, "#%d", match_thing(privs, args[0], MATCH_EVERYTHING));
}

FUNCTION(pnum) {
  dbref thing=lookup_player(args[0]);
  sprintf(buff, "#%d", (thing == AMBIGUOUS) ? NOTHING : thing);
}

FUNCTION(name) {
  INIT_MATCH(it, args[0]);
  strcpy(buff, it==NOTHING?"":db[it].name);
}

FUNCTION(fullname) {
  INIT_MATCH(it, args[0]);
  char *r, *s;

  if(it == NOTHING)
    *buff='\0';
  else {
    strcpy(buff, db[it].name);
    if(Typeof(it) == TYPE_EXIT && *(s=atr_get(it, A_ALIAS))) {
      r=strnul(buff);
      *r++=';';
      strmaxcpy(r, s, 8000-(r-buff));
    }
  }
}

FUNCTION(flags) {
  INIT_MATCH(it, args[0]);
  strcpy(buff, it==NOTHING?"#-1":unparse_flags(privs, it, 0));
}

FUNCTION(unparse) {
  INIT_MATCH(thing, args[0]);

  if(thing == NOTHING && *args[0] == '#')
    thing=atoi(args[0]+1);
  strcpy(buff, unparse_object(doer, thing));
}

FUNCTION(owner) {
  ATTR *attr;
  dbref thing;
  char *s;

  /* return owner of object */
  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) != NOTHING) {
        if(Typeof(thing) == TYPE_PLAYER || !HIDDEN_ADMIN ||
           class(db[thing].owner) != CLASS_WIZARD ||
           controls(privs, thing, POW_FUNCTIONS))
          thing=db[thing].owner;
        else
          thing=NOTHING;
      }
      sprintf(buff, "#%d", thing);
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  /* Find object to which an attribute is defined upon.
   * A #-1 means it is a builtin attribute. */

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING ||
     !(attr=atr_str(thing, s)))
    strcpy(buff, "#-1");
  else if(can_see_atr(privs, thing, attr))
    sprintf(buff, "#%d", attr->obj);  /* note: not owner of object! */
  else
    strcpy(buff, ERR_PERM);
}

static void session_data(char *buff, dbref privs, char *name, int type)
{
  INIT_CMATCH(thing, name, POW_FUNCTIONS);

  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else if(Typeof(thing) != TYPE_PLAYER)
    strcpy(buff, ERR_PLAYER);
  else
    sprintf(buff, "%d", !type ? db[thing].data->sessions:
            type==1?db[thing].data->steps:db[thing].data->age);
}

FUNCTION(sessions) { session_data(buff, privs, args[0], 0); }
FUNCTION(steps)    { session_data(buff, privs, args[0], 1); }
FUNCTION(age)      { session_data(buff, privs, args[0], 2); }

FUNCTION(cols) {
  INIT_MATCH(player, args[0]);
  
  if(player == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", get_cols(NULL, player));
}

FUNCTION(rows) {
  INIT_MATCH(player, args[0]);
  
  if(player == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", get_rows(NULL, player));
}

FUNCTION(term) {
  INIT_MATCH(player, args[0]);
  int i;

  if(player == NOTHING) {
    strcpy(buff, "#-1");
    return;
  }
  player=db[player].owner;

  for(i=0;term_config[i].name;i++)
    if(string_prefix(term_config[i].name, args[1]))
      break;

  if(!term_config[i].name)
    strcpy(buff, "#-1 No such term setting.");
  else {
    switch(i) {
    case 0: sprintf(buff, "%d", get_rows(NULL, player)); break;
    case 1: sprintf(buff, "%d", get_cols(NULL, player)); break;
    default:
      sprintf(buff, "%d", (db[player].data->term & term_config[i].flag)?1:0);
    }
  }
}

FUNCTION(type) {
  INIT_MATCH(thing, args[0]);
  strcpy(buff, (thing == NOTHING)?"#-1":typenames[Typeof(thing)]);
}

FUNCTION(class) {
  INIT_MATCH(thing, args[0]);
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else if(Typeof(thing) == TYPE_PLAYER) {
    if(HIDDEN_ADMIN && !Immortal(privs))
      strcpy(buff, ERR_PERM);
    else
      strcpy(buff, classnames[class(thing)]);
  } else
    strcpy(buff, typenames[Typeof(thing)]);
}

FUNCTION(rank) {
  INIT_MATCH(thing, args[0]);
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", class(db[thing].owner));
}

FUNCTION(immortal) {
  INIT_MATCH(thing, args[0]);
  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", Immortal(thing));
}

FUNCTION(gender) {
  INIT_MATCH(it, args[0]);

  char *types[4]={"Male", "Female", "Neuter", "Plural"};

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else
    strcpy(buff, types[get_gender(it)]);
}

static void gender_substitute(char *buff, dbref privs, char *name, int type)
{
  INIT_MATCH(it, name);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else
    strcpy(buff, gender_subst(it, type));
}

FUNCTION(subj)  { gender_substitute(buff, privs, args[0], SUBJ); }
FUNCTION(objn)  { gender_substitute(buff, privs, args[0], OBJN); }
FUNCTION(poss)  { gender_substitute(buff, privs, args[0], POSS); }
FUNCTION(aposs) { gender_substitute(buff, privs, args[0], APOSS); }

FUNCTION(rmatch) {
  INIT_CMATCH(who, args[0], POW_FUNCTIONS);
  if(who == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "#%d", match_thing(who, args[1], MATCH_EVERYTHING));
}

FUNCTION(nearby) {
  INIT_MATCH(obj1, args[0]);
  INIT_MATCH(obj2, args[1]);
  sprintf(buff, "%d", (locatable(privs, obj1) && locatable(privs, obj2) &&
          nearby(obj1, obj2)));
}

FUNCTION(sees) {
  INIT_MATCH(obj1, args[0]);
  INIT_MATCH(obj2, args[1]);

  if(obj1 == NOTHING || obj2 == NOTHING)
    strcpy(buff, "#-1");
  else if((!controls(privs, obj1, POW_FUNCTIONS) && !nearby(privs, obj1)) ||
          (!controls(privs, obj2, POW_FUNCTIONS) && !nearby(privs, obj2)))
    strcpy(buff, ERR_LOC);
  else
    sprintf(buff, "%d", can_see(obj1, obj2));
}

FUNCTION(loc) {
  INIT_MATCH(it, args[0]);
  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!locatable(privs, it))
    strcpy(buff, ERR_LOC);
  else
    sprintf(buff, "#%d", db[it].location);
}

FUNCTION(room) {
  INIT_MATCH(it, args[0]);
  int dep, var=it;

  if(it == NOTHING) {
    strcpy(buff, "#-1");
    return;
  } if(!locatable(privs, it)) {
    strcpy(buff, ERR_LOC);
    return;
  }

  for(dep=0;dep < 100 && var != db[var].location;dep++) {
    var=db[var].location;
    if(Invalid(var))
      break;
  }

  /* No room found? */
  if(!Invalid(var) && Typeof(var) != TYPE_ROOM)
    var=NOTHING;

  sprintf(buff, "#%d", var);
}

FUNCTION(rloc) {
  INIT_MATCH(it, args[0]);
  int dep, var=it, limit=atoi(args[1]);

  if(it == NOTHING) {
    strcpy(buff, "#-1");
    return;
  } if(!locatable(privs, it)) {
    strcpy(buff, ERR_LOC);
    return;
  }

  /* set a reasonable limit to the recursion */
  if(limit > 100)
    limit=100;

  for(dep=0;dep < limit && var != db[var].location;dep++) {
    var=db[var].location;
    if(Invalid(var))
      break;
  }

  sprintf(buff, "#%d", var);
}

FUNCTION(con) {
  INIT_MATCH(it, args[0]);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    DOLIST(it, db[it].contents)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        break;
    sprintf(buff, "#%d", it);
  }
}

FUNCTION(lcon) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[it].contents)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        ADDWORD(buff, s, "#%d", it);
  }
}

FUNCTION(plcon) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[it].contents)
      if(Typeof(it) == TYPE_PLAYER && (controls(privs, it, POW_FUNCTIONS) ||
         can_see(privs, it)))
        ADDWORD(buff, s, "#%d", it);
  }
}

FUNCTION(exit) {
  INIT_MATCH(it, args[0]);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    DOLIST(it, db[it].exits)
      if(controls(privs, it, POW_FUNCTIONS) || can_see(privs, it))
        break;
    sprintf(buff, "#%d", it);
  }
}

FUNCTION(lexits) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) && !nearby(privs, it))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[it].exits)
      if(controls(privs, it, POW_EXAMINE) || can_see(privs, it))
        ADDWORD(buff, s, "#%d", it);
  }
}

FUNCTION(randexit) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);
  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "#%d", random_exit(thing, NOTHING));
}

FUNCTION(next) {
  INIT_MATCH(it, args[0]);

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if((!controls(privs, it, POW_FUNCTIONS) &&
     !controls(privs, db[it].location, POW_FUNCTIONS) &&
     !nearby(privs, db[it].location)))
    strcpy(buff, ERR_LOC);
  else {
    DOLIST(it, db[it].next)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        break;
    sprintf(buff, "#%d", it);
  }
}

FUNCTION(objlist) {
  INIT_MATCH(it, args[0]);
  char *s=buff;

  if(it == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, it, POW_FUNCTIONS) &&
     !controls(privs, db[it].location, POW_FUNCTIONS) &&
     !nearby(privs, db[it].location))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, it)
      if((controls(privs, it, POW_FUNCTIONS) || can_see(privs, it)) &&
         (!IS(it, TYPE_THING, THING_MONSTER) || power(privs, POW_COMBAT)))
        ADDWORD(buff, s, "#%d", it);
  }
}

FUNCTION(link) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "#%d", db[it].link);
}

FUNCTION(linkup) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  int haspow;
  dbref i;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    haspow=controls(privs, thing, POW_FUNCTIONS);
    for(i=0;i<db_top;i++)
      if(db[i].link == thing && (haspow || controls(privs, i, POW_FUNCTIONS)))
        ADDWORD(buff, s, "#%d", i);
  }
}

FUNCTION(entrances) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  int haspow;
  dbref i;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    haspow=controls(privs, thing, POW_FUNCTIONS);
    for(i=0;i<db_top;i++)
      if(db[i].link == thing && Typeof(i) == TYPE_EXIT &&
         (haspow || controls(privs, i, POW_FUNCTIONS)))
        ADDWORD(buff, s, "#%d", i);
  }
}

FUNCTION(oexit) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref it;

  if(thing == NOTHING || Typeof(thing) != TYPE_EXIT)
    strcpy(buff, "#-1");
  else if(!controls(privs, thing, POW_FUNCTIONS))
    strcpy(buff, ERR_PERM);
  else if(db[thing].link == NOTHING)
    *buff=0;
  else {
    it=(db[thing].link == HOME) ? db[privs].link : db[thing].link;

    *buff=0;
    DOLIST(it, db[it].exits)
      if(db[it].link == db[thing].location && (controls(privs, it,
         POW_FUNCTIONS) || can_see(privs, it)))
        ADDWORD(buff, s, "#%d", it);
  }
}

FUNCTION(powers) {
  INIT_MATCH(player, args[0]);

  if(player == NOTHING)
    strcpy(buff, "#-1");
  else if(RESTRICT_HASPOW && !controls(privs, player, POW_EXAMINE))
    strcpy(buff, ERR_PERM);
  else
    disp_powers(buff, db[player].owner);
}

FUNCTION(controls) {
  INIT_MATCH(player, args[0]);
  INIT_MATCH(object, args[1]);
  int k;

  /* Determine power to check */
  if(nargs == 3) {
    for(k=0;k<NUM_POWS;k++)
      if(!strcasecmp(powers[k].name, args[2]))
        break;
  } else
    k=POW_MODIFY;

  if(player == NOTHING || object == NOTHING)
    strcpy(buff, "#-1");
  else if(k == NUM_POWS)
    strcpy(buff, "#-1 No such power.");
  else if(nargs == 3 && RESTRICT_HASPOW && (!controls(privs, player,
          POW_EXAMINE) || !controls(privs, object, POW_EXAMINE)))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", controls(player, object, k));
}

FUNCTION(haspow) {
  INIT_MATCH(player, args[0]);
  int k;

  for(k=0;k<NUM_POWS;k++)
    if(!strcasecmp(powers[k].name, args[1]))
      break;

  if(player == NOTHING)
    strcpy(buff, "#-1");
  else if(k == NUM_POWS)
    strcpy(buff, "#-1 No such power.");
  else if(RESTRICT_HASPOW && !controls(privs, player, POW_EXAMINE))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", power(player, k));
}

FUNCTION(parents) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref *ptr;
  int haspow;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    haspow=controls(privs, thing, POW_FUNCTIONS);
    for(ptr=db[thing].parents;ptr && *ptr != NOTHING;ptr++)
      if(haspow || controls(privs, *ptr, POW_FUNCTIONS))
        ADDWORD(buff, s, "#%d", *ptr);
  }
}

FUNCTION(children) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref *ptr;
  int haspow;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    haspow=controls(privs, thing, POW_FUNCTIONS);
    for(ptr=db[thing].children;ptr && *ptr != NOTHING;ptr++)
      if(haspow || controls(privs, *ptr, POW_FUNCTIONS))
        ADDWORD(buff, s, "#%d", *ptr);
  }
}

FUNCTION(is_a) {
  INIT_MATCH(thing, args[0]);
  INIT_MATCH(parent, args[1]);

  if(thing == NOTHING || parent == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, thing, POW_FUNCTIONS) && !controls(privs, parent,
          POW_FUNCTIONS))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", is_a(thing, parent));
}

FUNCTION(has_a) {
  INIT_MATCH(room, args[0]);
  INIT_MATCH(parent, args[1]);
  char *s=buff;
  dbref it;

  if(room == NOTHING || parent == NOTHING)
    strcpy(buff, "#-1");
  else if(!controls(privs, room, POW_FUNCTIONS) && !nearby(privs, room))
    strcpy(buff, ERR_LOC);
  else {
    *buff=0;
    DOLIST(it, db[room].contents)
      if(is_a(it, parent) && (controls(privs, it, POW_FUNCTIONS) ||
         controls(privs, parent, POW_FUNCTIONS)))
        ADDWORD(buff, s, "#%d", it);
  }
}

FUNCTION(relay) {
  INIT_CMATCH(thing, args[0], POW_FUNCTIONS);
  char *s=buff;
  dbref *ptr;

  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else {
    *buff=0;
    for(ptr=db[thing].relay;ptr && *ptr != NOTHING;ptr++)
      ADDWORD(buff, s, "#%d", thing);
  }
}

FUNCTION(zone) {
  INIT_MATCH(thing, args[0]);
  char *s=buff;
  dbref *ptr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else {
    *buff=0;
    for(ptr=db[mainroom(thing)].zone;ptr && *ptr != NOTHING;ptr++)
      ADDWORD(buff, s, "#%d", *ptr);
  }
}

FUNCTION(inzone) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  char *s=buff;
  dbref i;

  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else {
    *buff=0;
    for(i=0;i<db_top;i++)
      if(Typeof(i) == TYPE_ROOM && inlist(db[i].zone, it))
        ADDWORD(buff, s, "#%d", i);
  }
}

FUNCTION(zwho) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  char *s=buff;
  dbref i;

  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else {
    *buff=0;
    for(i=0;i<db_top;i++)
      if(Typeof(i) == TYPE_PLAYER && inlist(db[mainroom(i)].zone, it))
        ADDWORD(buff, s, "#%d", i);
  }
}

FUNCTION(plane) {
  INIT_CMATCH(thing, args[0], POW_EXAMINE);

  if(thing == NOTHING || !power(privs, POW_DARK))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", db[thing].plane);
}


/* Set 8: Database Attributes ************************************************/

FUNCTION(v) {
  if(strlen(args[0]) > 1) {   /* not a null or 1-character string. */
    ATTR *attr;
    
    if(!(attr=atr_str(privs, args[0])))
      *buff='\0';
    else if(can_see_atr(privs, privs, attr))
      strcpy(buff, atr_get_obj(privs, attr->obj, attr->num));
    else
      *buff='\0';
    return;
  }

  switch(tolower(*args[0])) {
    case '0' ... '9': strcpy(buff, wptr[*args[0]-'0']); return;
    case 'n': strcpy(buff, db[doer].name); return;
    case '#': sprintf(buff, "#%d", doer); return;
    case '!': sprintf(buff, "#%d", privs); return;
    case 'u': strcpy(buff, db[privs].name); return;
    case 'l': sprintf(buff, "#%d", locatable(privs, doer) ?
                      db[doer].location:NOTHING); return;
    default: *buff='\0';
  }
}

FUNCTION(get) {
  ATTR *attr;
  dbref thing;
  char *s;

  /* backward compatibility.. get(obj/attr) */
  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      strcpy(buff, "#-1 No attribute name specified.");
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING)
    strcpy(buff, "#-1");
  else if(!(attr=atr_str(thing, s)))
    *buff='\0';
  else if(can_see_atr(privs, thing, attr))
    strcpy(buff, atr_get_obj(thing, attr->obj, attr->num));
  else
    strcpy(buff, ERR_PERM);
}

FUNCTION(zoneattr) {
  ATTR *attr;
  dbref thing, *zone;
  char *s;

  /* also allow zoneattr(obj/attr) */
  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      strcpy(buff, "#-1 No attribute name specified.");
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING) {
    strcpy(buff, "#-1");
    return;
  }

  /* we only check zones that we can see attributes on */
  for(zone=db[mainroom(thing)].zone;zone && *zone != NOTHING;zone++)
    if((attr=atr_str(*zone, s)) && can_see_atr(privs, *zone, attr)) {
      strcpy(buff, atr_get_obj(*zone, attr->obj, attr->num));
      return;
    }

  *buff='\0';
  return;
}

FUNCTION(isattr) {
  INIT_MATCH(thing, args[0]);
  ATTR *attr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", (attr=atr_str(thing, args[1])) &&
                         can_see_atr(privs, thing, attr));
}

FUNCTION(hasattr) {
  INIT_MATCH(thing, args[0]);
  ATTR *attr;

  if(thing == NOTHING)
    strcpy(buff, "#-1");
  else
    sprintf(buff, "%d", (attr=atr_str(thing, args[1])) &&
            can_see_atr(privs, thing, attr) &&
            *(atr_get_obj(thing, attr->obj, attr->num)));
}

FUNCTION(attropts) {
  ATTR *attr;
  dbref thing;
  char *s;

  if(nargs == 1) {
    if(!(s=strchr(args[0], '/'))) {
      strcpy(buff, "#-1 No attribute name specified.");
      return;
    }
    *s++='\0';
  } else
    s=args[1];

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING)
    strcpy(buff, "#-1");
  else if(!(attr=atr_str(thing, s)))
    *buff='\0';
  else if(can_see_atr(privs, thing, attr))
    strcpy(buff, unparse_attrflags(attr->flags));
  else
    strcpy(buff, ERR_PERM);
}

FUNCTION(hasflag) {
  INIT_MATCH(thing, args[0]);
  ATTR *attr;
  char *s=strchr(args[0], '/');
  int a;

  if((nargs == 2 && s) || nargs == 3) {
    if(nargs == 2)
      *s++='\0';
    else
      s=args[1];
  } else
    s=NULL;

  if((thing=match_thing(privs, args[0], MATCH_EVERYTHING)) == NOTHING) {
    strcpy(buff, "#-1");
    return;
  }

  /* search for a flag on a particular attribute */
  if(s) {
    for(a=0;a<16;a++)
      if(string_prefix(attr_options[a], args[nargs-1]))
        break;
    if(a == 16)
      strcpy(buff, "#-1 No such attribute option.");
    else if(!(attr=atr_str(thing, s)) || !can_see_atr(privs, thing, attr))
      strcpy(buff, "0");
    else
      sprintf(buff, "%d", (attr->flags & (1 << a)) ? 1:0);
    return;
  }

  /* search for a flag on an object */

  for(a=0;flaglist[a].name;a++)
    if((strlen(args[1]) == 1 && flaglist[a].flag == *args[1]) ||
       (strlen(args[1]) > 1 && string_prefix(flaglist[a].name, args[1])))
      break;
  if(!flaglist[a].name)
    strcpy(buff, "#-1 No such flag.");
  else
    /* Don't show 'Connected' flag if player is hidden! */
    if(flaglist[a].objtype == TYPE_PLAYER && flaglist[a].bits ==
       PLAYER_CONNECT && !(could_doit(privs, thing, A_HLOCK) ||
       controls(privs, thing, POW_WHO)))
      strcpy(buff, "0");
  else
    sprintf(buff, "%d",
      (!flaglist[a].bits && Typeof(thing) == flaglist[a].objtype) ||
      (flaglist[a].bits && (flaglist[a].objtype == NOTYPE ||
       flaglist[a].objtype == Typeof(thing)) &&
       (db[thing].flags & flaglist[a].bits)));
}

FUNCTION(lattr) {
  INIT_MATCH(obj, args[0]);
  char *s=buff;
  ALIST *a;
  ATTR *atr;

  *buff=0;
  for(a=db[obj].attrs;a;a=a->next)
    if((atr=atr_num(a->obj, a->num)) && can_see_atr(privs, obj, atr)) {
      if(s-buff > 7930) {
        strcpy(s, " ...");
        return;
      }
      if(s != buff)
        *s++=' ';
      strcpy(s, atr->name);
      s+=strlen(s);
    }
}

FUNCTION(lattrdef) {
  INIT_MATCH(it, args[0]);
  char *s=buff;
  ATRDEF *k;

  *buff=0;
  if((db[it].flags & VISIBLE) || controls(privs, it, POW_EXAMINE))
    for(k=db[it].atrdefs;k;k=k->next) {
      if(s-buff > 7930) {
        strcpy(s, " ...");
        return;
      }
      if(s != buff)
        *s++=' ';
      strcpy(s, k->atr.name);
      s+=strlen(s);
    }
}

FUNCTION(setq) {
  int a=atoi(args[0]);

  if(a < 0 || a > 9)
    strcpy(buff, ERR_ENV);
  else {
    strcpy(buff, "");
    strcpy(wptr[a], args[1]);
  }
}

FUNCTION(setr) {
  int a=atoi(args[0]);

  if(a < 0 || a > 9)
    strcpy(buff, ERR_ENV);
  else {
    strcpy(buff, args[1]);
    strcpy(wptr[a], args[1]);
  }
}

FUNCTION(s) { pronoun_substitute(buff, privs, doer, args[0]); }

FUNCTION(s_as) {
  INIT_CMATCH(newdoer, args[1], POW_MODIFY);
  INIT_CMATCH(newprivs, args[2], POW_MODIFY);

  if(newdoer == NOTHING || newprivs == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    pronoun_substitute(buff, newprivs, newdoer, args[0]);
}

FUNCTION(s_as_with) {
  INIT_CMATCH(newdoer, args[1], POW_MODIFY);
  INIT_CMATCH(newprivs, args[2], POW_MODIFY);

  char *sptr[10];
  int a;

  if(newdoer == NOTHING || newprivs == NOTHING) {
    strcpy(buff, ERR_PERM);
    return;
  }

  /* preserve local registers */
  for(a=0;a<10;a++) {
    SPTR(sptr[a], wptr[a]);
    strcpy(wptr[a], (a < nargs-3)?args[a+3]:"");
  }

  pronoun_substitute(buff, newprivs, newdoer, args[0]);

  /* restore local registers */
  for(a=0;a<10;a++) {
    strcpy(wptr[a], sptr[a]);
    free(sptr[a]);
  }
}

FUNCTION(parse_lock) {
  INIT_MATCH(player, args[0]);
  INIT_CMATCH(thing, args[1], POW_EXAMINE);

  if(player == NOTHING || thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", eval_boolexp(player, thing, args[2]));
}

FUNCTION(createtime) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%ld", db[it].create_time);
}

FUNCTION(modtime) {
  INIT_CMATCH(it, args[0], POW_FUNCTIONS);
  if(it == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%ld", db[it].mod_time);
}

FUNCTION(pennies) {
  INIT_CMATCH(who, args[0], POW_MONEY);
  if(who == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", db[who].pennies);
}

FUNCTION(stats) {
  dbref owner=AMBIGUOUS;
  int total, obj[8], pla[NUM_CLASSES];
  
  if(nargs) {
    if((owner=lookup_player(args[0])) == NOTHING) {
      strcpy(buff, ERR_PLAYER);
      return;
    }

    if(owner == AMBIGUOUS)
      owner=db[privs].owner;
    else if(!controls(privs, owner, POW_STATS)) {
      strcpy(buff, ERR_PERM);
      return;
    }
  }

  calc_stats(owner, &total, obj, pla);
  sprintf(buff, "%d %d %d %d %d %d %d %d %d %d", db_top, total,
          obj[0], obj[1], obj[2], obj[3], obj[4], obj[5], obj[6], obj[7]);
}

FUNCTION(objmem) {
  INIT_CMATCH(thing, args[0], POW_STATS);
  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", mem_usage(thing));
}

FUNCTION(playmem) {
  INIT_CMATCH(thing, args[0], POW_STATS);
  if(thing == NOTHING)
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "%d", playmem_usage(db[thing].owner));
}


/* Set 9: String Utilities **************************************************/

FUNCTION(strlen) { sprintf(buff, "%d", (int)strlen(args[0])); }
FUNCTION(ascii)  { sprintf(buff, "%d", (unsigned char)*args[0]); }
FUNCTION(match)  { sprintf(buff, "%d", wild_match(args[0], args[1], 1, 0)); }
FUNCTION(edit)   { edit_string(buff, args[0], args[1], args[2]); }
FUNCTION(secure) { strcpy(buff, secure_string(args[0])); }

FUNCTION(comp) {
  int i=strcmp(args[0], args[1]);
  sprintf(buff, "%d", (i>0)?1 : (i<0)?-1 : 0);
}

FUNCTION(strcat) {
  char *s=buff;
  int i;

  *buff='\0';
  for(i=0;i<nargs && (s-buff < 8000);i++)
    strmaxcpy(s+=strlen(s), args[i], 8001-(s-buff));
}

FUNCTION(wmatch) {
  DELIM(delim, 2);
  char *r, *t=args[0];
  int wcount=1;

  while((r=parse_up(&t, delim))) {
    if(!*r && delim == ' ')
      continue;
    if(wild_match(r, args[1], 1, 0)) {
      sprintf(buff, "%d", wcount);
      return;
    }
    wcount++;
  }
  strcpy(buff, "0");
}

FUNCTION(matchall) {
  DELIM(delim, 2);
  char *r, *s=buff, *t=args[0];
  int wcount=1;

  *buff=0;
  while((r=parse_up(&t, delim))) {
    if(!*r && delim == ' ')
      continue;
    if(wild_match(r, args[1], 1, 0))
      ADDWORD(buff, s, "%d", wcount);
    wcount++;
  }
}
  
FUNCTION(regmatch) {
  char *sptr[10];
  int a=(nargs==3)?atoi(args[2]):0, b, result;

  if(a < 0 || a > 9) {
    strcpy(buff, ERR_ENV);
    return;
  }

  /* preserve copy of old environment */
  for(b=a;b<10;b++)
    SPTR(sptr[b], wptr[b]);

  if(!(result=wild_match(args[0], args[1], 0, a+1)))
    for(b=a;b<10;b++)
      strcpy(wptr[b], sptr[b]);  /* restore old environment if case failed */

  for(b=a;b<10;b++)
    free(sptr[b]);
  sprintf(buff, "%d", result);
}

FUNCTION(capstr) {
  *buff=toupper(*args[0]);
  if(*args[0])
    strcpy(buff+1, args[0]+1);
}

FUNCTION(ucstr) {
  char *s=buff, *t=args[0];

  while(*t)
    *s++=toupper(*t++);
  *s='\0';
}

FUNCTION(lcstr) {
  char *s=buff, *t=args[0];

  while(*t)
    *s++=tolower(*t++);
  *s='\0';
}

FUNCTION(pos) {
  char *s=strstr(args[1], args[0]);
  sprintf(buff, "%d", s?(int)(s-args[1]+1):0);
}

FUNCTION(lpos) {
  DELIM(ch, 1);
  char *r=args[0]-1, *s=buff;

  *buff=0;
  while((r=strchr(r+1, ch)))
    ADDWORD(buff, s, "%d", (int)(r-args[0]));
}

FUNCTION(mid) {
  int l=atoi(args[1]), len=atoi(args[2]);

  if(l < 0 || len < 0)
    strcpy(buff, ERR_RANGE);
  else if(l >= strlen(args[0]))
    *buff='\0';
  else {
    if(len > 8000)
      len=8000;
    strmaxcpy(buff, args[0]+l, len+1);  /* no buffer overrun */
  }
}

FUNCTION(delete) {
  int len=atoi(args[1]), l=atoi(args[2]);

  if(len < 0 || l < 0)
    strcpy(buff, ERR_RANGE);
  else {
    if(len > 8000)
      len=8000;
    strmaxcpy(buff, args[0], len+1);
    if(len+l < strlen(args[0]))
      strcpy(buff+len, args[0]+len+l);
  }
}

FUNCTION(first) {
  DELIM(delim, 1);
  char *r, *s=args[0];

  if(delim == ' ')
    while(*s && *s == ' ')
      s++;
  if((r=strchr(s, delim)))
    *r='\0';
  strcpy(buff, s);
}

FUNCTION(rest) {
  DELIM(delim, 1);
  char *s=args[0];

  if(delim == ' ')
    while(*s && *s == ' ')
      s++;
  if((s=strchr(s, delim))) {
    if(delim == ' ')
      while(*(s+1) == ' ')
        s++;
    strcpy(buff, s+1);
  } else
    *buff='\0';
}

FUNCTION(last) {
  DELIM(delim, 1);
  char *s=strnul(args[0]), *end=s;

  if(delim == ' ') {
    while(s > args[0] && *(s-1) == ' ')
      s--;
    end=s;
  }
  while(s > args[0] && *(s-1) != delim)
    s--;

  memcpy(buff, s, end-s);
  buff[end-s]='\0';
}

FUNCTION(extract) {
  DELIM(delim, 3);
  char *r, *s=args[0];
  int first=atoi(args[1]), last=first+atoi(args[2])-1, wcount=1;

  /* check length specified */
  if(last < first-1) {
    strcpy(buff, ERR_RANGE);
    return;
  } if(last < 1 || last < first) {
    *buff='\0';
    return;
  }

  if(delim == ' ')
    trim_spaces(&s);
  r=s;

  while((r=strchr(r, delim))) {
    if(wcount >= last)
      break;
    r++;
    if(delim == ' ')
      while(*r == delim)
        r++;
    if(!*r || ++wcount == first)
      s=r;
  }

  if(wcount < first)
    *buff=0;
  else {
    if(r)
      *r='\0';
    strcpy(buff, s);
  }
}

FUNCTION(remove) {
  DELIM(delim, 3);
  int first=atoi(args[1]), last=first+atoi(args[2]), wcount=1;
  char *r, *s=args[0];

  /* check length specified */
  if(last < first) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  if(delim == ' ')
    trim_spaces(&s);

  if(last <= 1 || last <= first) {
    strcpy(buff, s);
    return;
  }

  *buff='\0'; r=s;
  while((s=strchr(s, delim))) {
    if(++wcount == first)
      strmaxcpy(buff, r, s-r+1);
    s++;
    if(delim == ' ')
      while(*s == ' ')
        s++;
    if(wcount == last) {
      if((r=strnul(buff)) > buff)
        *r++=delim;
      strcpy(r, s);
      return;
    }
  }

  if(wcount < first)
    strcpy(buff, r);
}

FUNCTION(insert) {
  DELIM(delim, 3);
  char *p=buff, *r, *s=args[0];
  int first=atoi(args[1]), wcount=1;

  if(delim == ' ')
    trim_spaces(&s);
  r=s;

  /* scan for location to begin insert */
  if(first > 1) {
    while((s=strchr(s, delim))) {
      if(delim == ' ')
        while(*(s+1) == ' ')
          s++;
      if(++wcount == first) {
        *s++='\0';
        break;
      }
      s++;
    }

    strcpy(buff, r);
    p+=strlen(p);
    *p++=delim;
  }

  strmaxcpy(p, args[2], 8001-(p-buff));
  p+=strlen(p);

  /* copy remainder of string to result buffer */
  if(s && *s && p-buff < 8000) {
    *p++=delim;
    strmaxcpy(p, s, 8001-(p-buff));
  }
}

FUNCTION(replace) {
  DELIM(delim, 3);
  char *p=buff, *r, *s=args[0];
  int replace=atoi(args[1]), wcount=1;

  if(delim == ' ')
    trim_spaces(&s);
  r=s;

  /* scan for word to replace */
  if(replace > 1) {
    while((s=strchr(s, delim))) {
      if(delim == ' ')
        while(*(s+1) == ' ')
          s++;
      if(++wcount == replace) {
        *s++='\0';
        break;
      }
      s++;
    }

    strcpy(buff, r);
    p+=strlen(p);
    *p++=delim;
  }

  strmaxcpy(p, args[2], 8001-(p-buff));
  p+=strlen(p);

  /* copy remainder of string to result buffer */
  if(s && *s && p-buff < 8000) {
    if(replace < 1)
      *p++=delim;
    else
      s=strchr(s, delim);

    if(s)
      strmaxcpy(p, s, 8001-(p-buff));
  }
}

FUNCTION(swap) {
  DELIM(delim, 3);
  char *r, *s=args[0], *t;
  char *p1=NULL, *p2=NULL, *end1=NULL;
  int first=atoi(args[1]), second=atoi(args[2]), wcount;

  if(delim == ' ')
    trim_spaces(&s);

  /* check range */
  if(first == second || (first <= 1 && second <= 1)) {
    strcpy(buff, s);  /* nothing to do */
    return;
  } if(first > second) {  /* force first < second */
    int tmp=first;
    first=second;
    second=tmp;
  }

  /* Find both strings to replace. p1=str1, p2=str2, end1=end of str1. */
  for(r=s,wcount=1;*r;wcount++) {
    /* beginning of next word */
    if(wcount >= first && first > 0 && !p1)
      p1=r;
    else if(wcount == second)
      p2=r;

    while(*r && *r != delim)
      r++;
    if(p1 && !end1)
      end1=r;
    if(p2 || !*r)  /* at end of second word */
      break;

    r++;
    if(delim == ' ')
      while(*r == ' ')
        r++;
  }
  if(!p2 && wcount == second)  /* blank word at end of list */
    p2=r;

  /* Do nothing if 1st word wasn't found yet, or both words are out of range */
  if(first >= wcount || (!p1 && !p2)) {
    strcpy(buff, s);
    return;
  }

  /* copy up to string 1 & replace */
  t=buff;
  if(p1) {
    memcpy(t, s, p1-s);
    t+=p1-s;
    if(p2) {
      memcpy(t, p2, r-p2);
      t+=r-p2;
    } else {
      /* 2nd word did not exist. Advance end1 past delimiter. */
      if(*end1)
        end1++;
      if(delim == ' ')
        while(*end1 == ' ')
          end1++;
    }
  } else {
    /* prepend word */
    memcpy(t, p2, r-p2);
    t+=r-p2;
    *t++=delim;
  }

  /* if first word doesn't exist, set end1 to beginning of string */
  if(!end1)
    end1=s;

  /* move word to end if second word wasn't found */
  if(!p2) {
    strcpy(t, end1);
    t+=strlen(t);

    /* append word */
    *t++=delim;
    memcpy(t, p1, end1-p1);
    t+=end1-p1;

    /* trim delimiter */
    if(*(t-1) == delim)
      t--;
    if(delim == ' ')
      while(*(t-1) == ' ')
        t--;
    *t='\0';
    return;
  }

  /* else copy up to string 2 & replace */
  memcpy(t, end1, p2-end1);
  t+=p2-end1;
  if(!p1 && !*r) {
    /* back out delimiter */
    t--;
    if(delim == ' ')
      while(*(t-1) == ' ')
        t--;
  } else if(!p1) {
    /* 1st word did not exist. Advance 'r' (end2) past delimiter. */
    r++;
    if(*r == ' ')
      while(*r == ' ')
        r++;
  } else {
    /* insert word */
    memcpy(t, p1, end1-p1);
    t+=end1-p1;
  }

  /* copy end of string */
  strcpy(t, r);
}

FUNCTION(wcount) {
  DELIM(delim, 1);
  char *s=args[0];
  int wcount=1;
  
  if(!*args[0]) {
    strcpy(buff, "0");
    return;
  }

  if(delim == ' ')
    trim_spaces(&s);

  while((s=strchr(s, delim))) {
    wcount++;
    s++;
    if(delim == ' ')
      while(*s == ' ')
        s++;
  }

  sprintf(buff, "%d", wcount);
}

FUNCTION(flip) {
  char *s=strnul(args[0]), *r=buff;

  while(s > args[0])
    *r++=*--s;
  *r='\0';
}

FUNCTION(wordflip) {
  DELIM(delim, 1);
  char *r=buff, *s=args[0], *t=strnul(args[0]);

  /* remove leading and following spaces */
  if(delim == ' ') {
    while(*s && *s == ' ')
      s++;
    while(t > s && *(t-1) == ' ')
      t--;
  }

  *buff=0;
  while(t >= s) {
    *t--='\0';

    while(t >= s && *t != delim)
      t--;
    if(*buff)
      *r++=delim;
    strcpy(r, t+1);
    r+=strlen(r);
  }
}

FUNCTION(scramble) {
  int i, j, n=strlen(args[0]);
  char c;

  strcpy(buff, args[0]);

  for(i=0;i<n;i++) {
    j=rand_num(n);
    c=buff[i];
    buff[i]=buff[j];
    buff[j]=c;
  }
}

FUNCTION(rand) { sprintf(buff, "%d", rand_num(atoi(args[0]))); }

FUNCTION(dice) {
  int total, count, n=atoi(args[0]), die=atoi(args[1]);

  if(n < 0 || n > 100)
    strcpy(buff, ERR_RANGE);
  else {
    for(total=0,count=0;count < n;count++)
      total += rand_num(die);
    sprintf(buff, "%d", total+n);
  }
}

FUNCTION(randword) {
  DELIM(delim, 1);
  char *s=args[0], *t=args[0];
  int wcount=1, num;
  
  if(!*args[0]) {
    *buff='\0';
    return;
  }

  while((s=strchr(s, delim))) {
    wcount++;
    s++;
    if(delim == ' ')
      while(*s == ' ')
        s++;
  }

  num=rand_num(wcount);
  wcount=0;
  s=args[0];

  while((t=strchr(t, delim))) {
    if(wcount >= num)
      break;
    t++;
    if(delim == ' ')
      while(*t == delim)
        t++;
    if(!*t || ++wcount == num)
      s=t;
  }

  if(t)
    *t='\0';
  strcpy(buff, s);
}

FUNCTION(base) {
  int digit, neg=0, oldbase=atoi(args[1]), newbase=atoi(args[2]);
  char temp[66], *r=args[0], *s;
  unsigned quad decimal=0;

  if(oldbase < 2 || oldbase > 36 || newbase < 2 || newbase > 36) {
    strcpy(buff, "#-1 Bases must be between 2 and 36.");
    return;
  }
  
  /* convert from oldbase */
  while(*r == ' ')
    r++;
  if(*r == '-') {
    neg=1;
    r++;
  } else if(*r == '+')
    r++;

  while(*r && *r != ' ') {
    digit=(*r >= '0' && *r <= '9')?(*r-48):((*r|32) >= 'a' && (*r|32) <= 'z')?
           ((*r|32)-87):36;
    if(digit >= oldbase) {
      strcpy(buff, "#-1 Illegal digit.");
      return;
    }
    decimal*=oldbase;
    decimal+=digit;
    r++;
  }

  if(!decimal) {
    strcpy(buff, "0");
    return;
  }

  /* convert to newbase (backwards) */
  s=temp+65;
  *s='\0';

  while(decimal > 0) {
    digit = decimal % newbase;
    decimal /= newbase;
    *--s = digit+(digit < 10?48:87);
  }

  if(neg)
    *--s='-';
  strcpy(buff, s);
}

FUNCTION(chr) {
  int chr=atoi(args[0]);

  if(!power(privs, POW_ANIMATION) && (chr < 32 || chr >= 127))
    chr=' ';
  buff[0]=chr;
  buff[1]='\0';
}

FUNCTION(esc) {
  if(!power(privs, POW_ANIMATION))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "\e[%s", (nargs == 1)?args[0]:"");
}

FUNCTION(csi) {
  if(!power(privs, POW_ANIMATION))
    strcpy(buff, ERR_PERM);
  else
    sprintf(buff, "\e%s", (nargs == 1)?args[0]:"");
}

FUNCTION(ansi) {
  int color=color_code(7, args[0]);

  if(color & 1024)
    strcpy(buff, rainbowmsg(args[1], color, 8000));
  else
    sprintf(buff, "%s%s%s", textcolor(default_color, color), args[1],
            textcolor(color, default_color));
}

FUNCTION(strip) { strcpy(buff, args[0]); }  // Stripped by F_STRIP flag.

static void increment_string(char *buff, char *str, int inc)
{
  char *s;

  if(!*str) {
    sprintf(buff, "%d", inc);
    return;
  }
  s=strnul(str)-1;
  if(!isdigit(*s)) {
    strcpy(buff, "#-1 String must end in an integer.");
    return;
  }

  while(s > str && (isdigit(*(s-1)) || *(s-1) == '-')) {
    s--;
    if(*s == '-')
      break;
  }

  memcpy(buff, str, s-str);
  sprintf(buff+(s-str), "%d", atoi(s)+inc);
}

FUNCTION(inc) { increment_string(buff, args[0], 1); }
FUNCTION(dec) { increment_string(buff, args[0], -1); }

FUNCTION(spc) {
  int len=atoi(args[0]);

  if(len < 1)
    *buff='\0';
  else {
    if(len > 8000)
      len=8000;
    memset(buff, ' ', len);
    buff[len]='\0';
  }
}

FUNCTION(lnum) {
  int a, i=(nargs >= 2)?atoi(args[1]):0, x=atoi(args[0]);
  int j=(nargs == 3)?atoi(args[2]):1;
  char *s=buff;

  *buff=0;
  for(a=0;a<x;a++)
    ADDWORD(buff, s, "%d", i+a*j);
}

FUNCTION(left) {
  int len=atoi(args[1]);

  if(len < 1)
    *buff='\0';
  else if(len > 7999)
    strcpy(buff, args[0]);
  else
    strmaxcpy(buff, args[0], len+1);
}

FUNCTION(right) {
  int len=atoi(args[1]);

  if(len < 1)
    *buff='\0';
  else {
    if((len=strlen(args[0])-len) < 0)
      len=0;
    strcpy(buff, args[0]+len);
  }
}

FUNCTION(ljust) {
  DELIM(fill, 2);
  int i=strlen(args[0]), len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  /* if string is smaller than length, pad with fill character */
  if(len > i) {
    memcpy(buff, args[0], i);
    memset(buff+i, fill, len-i);
    buff[len]='\0';
  } else
    strmaxcpy(buff, args[0], len+1);
}

FUNCTION(rjust) {
  DELIM(fill, 2);
  int i=strlen(args[0]), len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  if(len > i) {
    memset(buff, fill, len-i);
    strcpy(buff+(len-i), args[0]);
  } else
    strcpy(buff, args[0]+i-len);
}

FUNCTION(center) {
  DELIM(fill, 2);
  int len=atoi(args[1]);

  if(len < 0 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  strcpy(buff, center(args[0], len, fill));
}

FUNCTION(trim) {
  DELIM(trim_char, 2);
  char *str=args[0], *end;

  /* trim left side of string */
  if(nargs < 2 || tolower(*args[1]) != 'r')
    while(*str && *str == trim_char)
      str++;

  /* trim right side */
  end=strnul(str);
  if(nargs < 2 || tolower(*args[1]) != 'l')
    while(end > str && *(end-1) == trim_char)
      end--;

  /* copy into result buffer */
  memcpy(buff, str, end-str);
  buff[end-str]='\0';
}

FUNCTION(justify) {
  int count=0, len=atoi(args[1]), line=atoi(args[2]);
  char *r, *s=args[0];

  if(len < 1 || len > 8000) {
    strcpy(buff, ERR_RANGE);
    return;
  } if(line < 0) {
    strcpy(buff, "#-1 Bad line number.");
    return;
  }

  /* If line=0, count number of lines for <len> justification,
     otherwise print the appropriate line requested. */

  while((r=eval_justify(&s, len)))
    if(++count == line)
      break;
  if(!line)
    sprintf(buff, "%d", count);
  else
    strcpy(buff, r?:"");
}

FUNCTION(switch) {
  int i;

  for(i=1;i+1 < nargs;i+=2)
    if(wild_match(args[0], args[i], 1, 0)) {
      strcpy(buff, args[i+1]);
      return;
    }

  if(i < nargs)
    strcpy(buff, args[i]);
  else
    *buff='\0';
}

FUNCTION(repeat) {
  char *s=buff;
  int a=atoi(args[1]);

  if(a < 0) {
    strcpy(buff, ERR_RANGE);
    return;
  }

  *buff=0;
  if(!*args[0])
    return;

  while(--a >= 0 && (s-buff) < 8000)
    strmaxcpy(s+=strlen(s), args[0], 8001-(s-buff));
}

FUNCTION(foreach) {
  DELIM(delim, 2);
  DELIMCH(odelim, 3, delim);
  char *p=buff, *r, *s=args[0], *sptr, temp[8192];

  *buff=0;
  if(!*args[0])
    return;

  SPTR(sptr, wptr[0]);

  while(func_total <= MAX_FUNCTIONS && (r=parse_perfect(&s, delim))) {
    strcpy(wptr[0], r);
    pronoun_substitute(temp, privs, doer, args[1]);
    if(!*temp)
      continue;
    if(p != buff) {
      if(p-buff+strlen(temp) > 8000)
        break;
      *p++=odelim;
    }
    strcpy(p, temp);
    p+=strlen(p);
  }

  strcpy(wptr[0], sptr);
  free(sptr);

  if(func_total > MAX_FUNCTIONS)
    strcpy(buff, "#-1 Too many functions.");
}

FUNCTION(oper) {
  DELIM(delim, 3);
  char *r, *s=args[0], *sptr[2];

  *buff=0;
  if(!*args[0]) {
    strcpy(buff, args[1]);
    return;
  }

  SPTR(sptr[0], wptr[0]);
  SPTR(sptr[1], wptr[1]);
  strcpy(wptr[1], args[1]);

  while(func_total < MAX_FUNCTIONS && (r=parse_perfect(&s, delim))) {
    strcpy(wptr[0], r);
    pronoun_substitute(buff, privs, doer, args[2]);
    strcpy(wptr[1], buff);
  }

  strcpy(wptr[0], sptr[0]);
  strcpy(wptr[1], sptr[1]);
  free(sptr[0]);
  free(sptr[1]);

  if(func_total > MAX_FUNCTIONS)
    strcpy(buff, "#-1 Too many functions.");
}


/* function flags */
#define F_EVAL	0x1	// Function arguments are automatically evaluated.
#define F_STRIP	0x2	// Strip ansi escape codes from arguments.

struct fun {
  char *name;
  void (*fun)();
  int flags:16;		// Function flags (listed above).
  unsigned int loarg:8;	// Minimum #args
  unsigned int hiarg:8;	// Maximum #args (100 = Unlimited)
};

static struct fun flist[]={
  {"ABS",	fun_abs,	0, 1},
  {"ACOS",	fun_acos,	0, 1},
  {"ADD",	fun_add,	0, 1,100},
  {"AGE",	fun_age,	0, 1},
  {"ANSI",	fun_ansi,	2, 2},
  {"APOSS",	fun_aposs,	0, 1},
  {"ASCII",	fun_ascii,	0, 1},
  {"ASIN",	fun_asin,	0, 1},
  {"ATAN",	fun_atan,	0, 1},
  {"ATTROPTS",	fun_attropts,	0, 1,2},
  {"AVG",	fun_avg,	0, 1,100},
  {"BAND",	fun_band,	0, 2},
  {"BASE",	fun_base,	2, 3},
  {"BETWEEN",	fun_between,	0, 3},
  {"BITCOUNT",	fun_bitcount,	0, 1},
  {"BNAND",	fun_bnand,	0, 2},
  {"BOR",	fun_bor,	0, 2},
  {"BXOR",	fun_bxor,	0, 2},
  {"CAPSTR",	fun_capstr,	2, 1},
  {"CEIL",	fun_ceil,	0, 1},
  {"CENTER",	fun_center,	2, 2,3},
  {"CHILDREN",	fun_children,	0, 1},
  {"CHR",	fun_chr,	0, 1},
  {"CLASS",	fun_class,	0, 1},
  {"CMDS",	fun_cmds,	0, 1},
  {"COLS",	fun_cols,	0, 1},
  {"COMP",	fun_comp,	2, 2},
  {"CON",	fun_con,	0, 1},
  {"CONCID",	fun_concid,	0, 1},
  {"CONFIG",	fun_config,	0, 1},
  {"CONTROLS",	fun_controls,	0, 2,3},
  {"COS",	fun_cos,	0, 1},
  {"CPU",	fun_cpu,	0, 0},
  {"CPUTIME",	fun_cputime,	0, 0},
  {"CREATETIME",fun_createtime,	0, 1},
  {"CREDITS",	fun_pennies,	0, 1},
  {"CSI",	fun_csi,	0, 0,1},
  {"DATE",	fun_date,	0, 0,1},
  {"DBTOP",	fun_dbtop,	0, 0},
  {"DEC",	fun_dec,	2, 1},
  {"DEGREES",	fun_degrees,	0, 2},
  {"DELETE",	fun_delete,	2, 3},
  {"DICE",	fun_dice,	0, 2},
  {"DIST2D",	fun_dist2d,	0, 4},
  {"DIST3D",	fun_dist3d,	0, 6},
  {"DIV",	fun_div,	0, 2},
  {"E",		fun_e,		0, 0},
  {"EDIT",	fun_edit,	2, 3},
  {"ENTRANCES",	fun_entrances,	0, 1},
  {"ESC",	fun_esc,	0, 0,1},
  {"EQ",	fun_eq,		0, 2},
  {"EXIT",	fun_exit,	0, 1},
  {"EXP",	fun_exp,	0, 1},
  {"EXTRACT",	fun_extract,	2, 3,4},
  {"FACTOR",	fun_factor,	0, 1},
  {"FADD",	fun_fadd,	0, 1,100},
  {"FDIV",	fun_fdiv,	0, 2},
  {"FFS",	fun_ffs,	0, 1},
  {"FIRST",	fun_first,	2, 1,2},
  {"FLAGS",	fun_flags,	0, 1},
  {"FLIP",	fun_flip,	2, 1},
  {"FLOOR",	fun_floor,	0, 1},
  {"FMUL",	fun_fmul,	0, 1,100},
  {"FOREACH",	fun_foreach,	2, 2,4},
  {"FSUB",	fun_fsub,	0, 2},
  {"FULLNAME",	fun_fullname,	0, 1},
  {"GENDER",	fun_gender,	0, 1},
  {"GET",	fun_get,	0, 1,2},
  {"GT",	fun_gt,		0, 2},
  {"GTEQ",	fun_gteq,	0, 2},
  {"HASATTR",	fun_hasattr,	0, 2},
  {"HASFLAG",	fun_hasflag,	0, 2,3},
  {"HASPOW",	fun_haspow,	0, 2},
  {"HAS_A",	fun_has_a,	0, 2},
  {"HOME",	fun_link,	0, 1},
  {"HOST",	fun_host,	0, 1},
  {"IDLE",	fun_idle,	0, 1},
  {"IF",	fun_if,		0, 2},
  {"IFELSE",	fun_ifelse,	0, 3},
  {"IFTRUE",	fun_iftrue,	0, 2},
  {"IMMORTAL",	fun_immortal,	0, 1},
  {"INC",	fun_inc,	2, 1},
  {"INPUT",	fun_input,	0, 1},
  {"INSERT",	fun_insert,	2, 3,4},
  {"INT",	fun_int,	0, 1},
  {"INZONE",	fun_inzone,	0, 1},
  {"ISATTR",	fun_isattr,	0, 2},
  {"IS_A",	fun_is_a,	0, 2},
  {"JUSTIFY",	fun_justify,	0, 3},
  {"LAND",	fun_land,	0, 2},
  {"LAST",	fun_last,	2, 1,2},
  {"LASTON",	fun_laston,	0, 1},
  {"LASTOFF",	fun_lastoff,	0, 1},
  {"LATTR",	fun_lattr,	0, 1},
  {"LATTRDEF",	fun_lattrdef,	0, 1},
  {"LCON",	fun_lcon,	0, 1},
  {"LCSTR",	fun_lcstr,	2, 1},
  {"LEFT",	fun_left,	2, 2},
  {"LEXITS",	fun_lexits,	0, 1},
  {"LINK",	fun_link,	0, 1},
  {"LINKUP",	fun_linkup,	0, 1},
  {"LJUST",	fun_ljust,	2, 2,3},
  {"LN",	fun_ln,		0, 1},
  {"LNUM",	fun_lnum,	0, 1,3},
  {"LOC",	fun_loc,	0, 1},
  {"LOG",	fun_log,	0, 1,2},
  {"LOR",	fun_lor,	0, 2},
  {"LPOS",	fun_lpos,	2, 2},
  {"LT",	fun_lt,		0, 2},
  {"LTEQ",	fun_lteq,	0, 2},
  {"LWHO",	fun_lwho,	0, 0},
  {"LXOR",	fun_lxor,	0, 2},
  {"MATCH",	fun_match,	2, 2},
  {"MATCHALL",	fun_matchall,	2, 2,3},
  {"MAX",	fun_max,	0, 1,100},
  {"MID",	fun_mid,	2, 3},
  {"MIN",	fun_min,	0, 1,100},
  {"MOD",	fun_mod,	0, 2},
  {"MODTIME",	fun_modtime,	0, 1},
  {"MONEY",	fun_pennies,	0, 1},
  {"MUDNAME",	fun_mudname,	0, 0},
  {"MUL",	fun_mul,	0, 1,100},
  {"NAME",	fun_name,	0, 1},
  {"NEARBY",	fun_nearby,	0, 2},
  {"NEG",	fun_neg,	0, 1},
  {"NEQ",	fun_neq,	0, 2},
  {"NETSTAT",	fun_netstat,	0, 1},
  {"NEXT",	fun_next,	0, 1},
  {"NOT",	fun_not,	0, 1},
  {"NUM",	fun_num,	0, 1},
  {"OBJLIST",	fun_objlist,	0, 1},
  {"OBJMEM",	fun_objmem,	0, 1},
  {"OBJN",	fun_objn,	0, 1},
  {"OEXIT",	fun_oexit,	0, 1},
  {"ONFOR",	fun_onfor,	0, 1},
  {"OPER",	fun_oper,	2, 3,4},
  {"OUTPUT",	fun_output,	0, 1},
  {"OWNER",	fun_owner,	0, 1,2},
  {"PARENTS",	fun_parents,	0, 1},
  {"PARSE_LOCK",fun_parse_lock,	0, 3},
  {"PENNIES",	fun_pennies,	0, 1},
  {"PI",	fun_pi,		0, 0},
  {"PLANE",	fun_plane,	0, 1},
  {"PLAYMEM",	fun_playmem,	0, 1},
  {"PLCON",	fun_plcon,	0, 1},
  {"PNUM",	fun_pnum,	0, 1},
  {"PORT",	fun_port,	0, 1},
  {"POS",	fun_pos,	2, 2},
  {"POSS",	fun_poss,	0, 1},
  {"POW",	fun_pow,	0, 2},
  {"POWERS",	fun_powers,	0, 1},
  {"RAND",	fun_rand,	0, 1},
  {"RANDEXIT",	fun_randexit,	0, 1},
  {"RANDWORD",	fun_randword,	2, 1,2},
  {"RANK",	fun_rank,	0, 1},
  {"REGMATCH",	fun_regmatch,	2, 2,3},
  {"RELAY",	fun_relay,	0, 1},
  {"REMOVE",	fun_remove,	2, 3,4},
  {"REPEAT",	fun_repeat,	0, 2},
  {"REPLACE",	fun_replace,	2, 3,4},
  {"REST",	fun_rest,	2, 1,2},
  {"RESTARTS",	fun_restarts,	0, 0},
  {"RIGHT",	fun_right,	2, 2},
  {"RJUST",	fun_rjust,	2, 2,3},
  {"RLOC",	fun_rloc,	0, 2},
  {"RMATCH",	fun_rmatch,	0, 2},
  {"ROOM",	fun_room,	0, 1},
  {"ROUND",	fun_round,	0, 1,2},
  {"ROWS",	fun_rows,	0, 1},
  {"RTIME",	fun_rtime,	0, 0,1},
  {"S",		fun_s,		0, 1},
  {"SCRAMBLE",	fun_scramble,	2, 1},
  {"SECURE",	fun_secure,	2, 1},
  {"SEES",	fun_sees,	0, 2},
  {"SESSIONS",	fun_sessions,	0, 1},
  {"SETQ",	fun_setq,	0, 2},
  {"SETR",	fun_setr,	0, 2},
  {"SHL",	fun_shl,	0, 2},
  {"SHR",	fun_shr,	0, 2},
  {"SIGN",	fun_sign,	0, 1},
  {"SIN",	fun_sin,	0, 1},
  {"SPC",	fun_spc,	0, 1},
  {"SQRT",	fun_sqrt,	0, 1},
  {"STATS",	fun_stats,	0, 0,1},
  {"STEPS",	fun_steps,	0, 1},
  {"STRCAT",	fun_strcat,	0, 2,100},
  {"STRIP",	fun_strip,	2, 1},
  {"STRLEN",	fun_strlen,	2, 1},
  {"SUB",	fun_sub,	0, 2},
  {"SUBJ",	fun_subj,	0, 1},
  {"SWAP",	fun_swap,	2, 3,4},
  {"SWITCH",	fun_switch,	0, 2,100},
  {"S_AS",	fun_s_as,	0, 3},
  {"S_AS_WITH",	fun_s_as_with,	0, 3,13},
  {"TAN",	fun_tan,	0, 1},
  {"TERM",	fun_term,	0, 2},
  {"TIME",	fun_time,	0, 0,1},
  {"TIMEOFDAY",	fun_timeofday,	0, 0},
  {"TMA",	fun_tma,	0, 1},
  {"TMF",	fun_tmf,	0, 1},
  {"TML",	fun_tml,	0, 1},
  {"TMS",	fun_tms,	0, 1},
  {"TRIM",	fun_trim,	2, 1,3},
  {"TRUTH",	fun_truth,	0, 1},
  {"TYPE",	fun_type,	0, 1},
  {"UCSTR",	fun_ucstr,	2, 1},
  {"UNPARSE",	fun_unparse,	0, 1},
  {"UPFRONT",	fun_upfront,	0, 0},
  {"UPTIME",	fun_uptime,	0, 0},
  {"V",		fun_v,		0, 1},
  {"VERSION",	fun_version,	0, 0},
  {"WCOUNT",	fun_wcount,	2, 1,2},
  {"WDATE",	fun_wdate,	0, 0,1},
  {"WMATCH",	fun_wmatch,	2, 2,3},
  {"WORDFLIP",	fun_wordflip,	2, 1,2},
  {"WTIME",	fun_wtime,	0, 0,1},
  {"XTIME",	fun_xtime,	0, 0,1},
  {"ZONE",	fun_zone,	0, 1},
  {"ZONEATTR",	fun_zoneattr,	0, 1,2},
  {"ZWHO",	fun_zwho,	0, 1},
  {0}};

void list_functions(dbref player)
{
  char buf[512];
  int a, col=get_cols(NULL, player);

  notify(player, "List of hardcoded functions in alphabetical order:");
  notify(player, "%s", "");
  strcpy(buf, "");
  for(a=0;flist[a].name;a++) {
    if(*buf && strlen(buf)+strlen(flist[a].name) >= col-2) {
      notify(player, "%s", buf);
      strcpy(buf, "");
    }
    if(*buf)
      strcat(buf, ", ");
    strcat(buf, flist[a].name);
  }
  if(*buf)
    notify(player, "%s", buf);
  notify(player, "%s", "");
}

/* search for a user-defined function */
static int user_defined(char **str, char *buff, dbref privs, dbref doer)
{
  ATTR *attr;
  char temp[8192], *saveptr[10], *writeptr[10], *s=buff;
  dbref thing=NOTHING, *ptr;
  int a, b;

  if(*s == '!')
    s++;

  /* check for function defined on a specific object.  e.g. #123.func() */
  if(*s == '#') {
    thing=atoi(s+1);
    if(Invalid(thing) || !(s=strchr(s, '.')))
      return 0;
    if(!(attr=atr_str(thing, s+1)) || !(attr->flags & AF_FUNC) ||
       !can_see_atr(privs, thing, attr))
      return 0;
  } else if((attr=atr_str(privs, s)) && (attr->flags & AF_FUNC))
    thing=privs;
  else {
    /* Scan zones for matching function */
    for(ptr=db[mainroom(privs)].zone;ptr && *ptr != NOTHING;ptr++)
      if((attr=atr_str(*ptr, s)) && (attr->flags & AF_FUNC)) {
        thing=*ptr;
        break;
      }

    /* If still nothing, try universal zones */
    if(thing == NOTHING)
      for(ptr=uzo;ptr && *ptr != NOTHING;ptr++)
        if((attr=atr_str(*ptr, s)) && (attr->flags & AF_FUNC)) {
          thing=*ptr;
          break;
        }
  }

  /* Nothing found. */
  if(thing == NOTHING)
    return 0;

  /* get the 10 arguments to the function */
  for(a=0;**str;a++) {
    if(a >= 10) {  /* Skip over remaining arguments */
      while(**str && **str != ')' && **str != ']')
        (*str)++;
      break;
    }
    if(**str == ')' || **str == ']') {
      SPTR(writeptr[a], "");
      continue;
    }
    evaluate(str, buff, privs, doer, 1);
    SPTR(writeptr[a], buff);
    if(**str == ',')
      (*str)++;
  }
  for(b=0;b<10;b++) {
    SPTR(saveptr[b], wptr[b]);
    strcpy(wptr[b], b<a?writeptr[b]:"");
    if(b<a)
      free(writeptr[b]);
  }

  if(**str)
    (*str)++;
  strcpy(temp, atr_get_obj(thing, attr->obj, attr->num));
  pronoun_substitute(buff, privs, doer, temp);

  for(b=0;b<10;b++) {
    strcpy(wptr[b], b<a?saveptr[b]:"");
    free(saveptr[b]);
  }
  return 1;
}


/* Exported functions */

void func_zerolev()
{
  func_recur=0; func_total=0;
}

char *func_printlev()
{
  return tprintf("%d %d", func_total, func_recur);
}

int function(char **str, char *buff, dbref privs, dbref doer)
{
  static hash *funchash;
  struct fun *fp;
  char **args=NULL, *s;
  int a, argc=0, inv=0;

  /* function names cannot have spaces or symbols in them */
  if(!ok_attribute_name(buff))
    return 0;

  if(++func_recur > FUNC_RECURSION) {
    strcpy(buff, "#-1 Too many levels of recursion.");
    func_recur--;
    return 1;
  } if(++func_total > MAX_FUNCTIONS) {
    strcpy(buff, "#-1 Too many functions.");
    func_recur--;
    return 1;
  }

  /* determine the function name we wish to find */
  s=buff;
  if(*s == '!') {
    inv=1;
    s++;
  }
  if(!funchash)
    funchash=make_hashtab(flist, sizeof(struct fun));
  fp=lookup_hash(funchash, s);

  /* Not a built-in function */
  if(!fp) {
    if(user_defined(str, buff, privs, doer)) {  // User-defined function found
      if(inv)
        sprintf(buff, "%d", !ISTRUE(buff));
      func_recur--;
      return 1;
    }
    func_recur--;
    return 0;
  }

  /* get the arguments to the function */
  if(**str != ')' && **str != ']')
    while(**str) {
      evaluate(str, buff, privs, doer, 1);

      if(!(argc%10))
        args=realloc(args, (argc+10)*sizeof(char *));  /* resize arg table */
      if(!args) {
        perror("argument allocation");
        argc=0;
        break;
      }

      SPTR(args[argc], buff);

      /* strip ansi escape codes from argument */
      if(fp->flags & F_STRIP)
        strcpy(args[argc], ansi_strip(args[argc]));

      /* check argument delimiters or end of function */
      argc++;
      if(**str == ',')
        (*str)++;
      else if(**str == ')' || **str == ']')
        break;
    }

  if(**str)
    (*str)++;

  if(argc < fp->loarg || (fp->hiarg < 100 && argc > (fp->hiarg?:fp->loarg))) {
    s=buff+sprintf(buff, "#-1 Function %s requires %s", fp->name,
                   fp->hiarg >= 100?"at least ":"");
    if(!fp->hiarg || fp->hiarg >= 100)
      sprintf(s, "%d argument%s.", fp->loarg, fp->loarg==1?"":"s");
    else if(fp->loarg == fp->hiarg-1)
      sprintf(s, "%d or %d arguments.", fp->loarg, fp->hiarg);
    else
      sprintf(s, "between %d and %d arguments.", fp->loarg, fp->hiarg);
  } else {
    make_quiet=1;
    fp->fun(buff, args, privs, doer, argc);
    make_quiet=0;
    if(inv)
      sprintf(buff, "%d", !ISTRUE(buff));
  }

  for(a=0;a<argc;a++)
    free(args[a]);
  free(args);
  func_recur--;
  return 1;
}

// Evaluates a function and returns its result.
//
// type=Delimiter character to end subcommand processing.
// type=1: Special; called from above to parse individual function arguments.
//
void evaluate(char **str, char *buff, dbref privs, dbref doer, char type)
{
  char *s=*str, *e=buff, *end=buff, temp[8192];
  int depth, chr, func=0;

  /* Skip preceding spaces */
  while(*s && *s == ' ')
    s++;

  while(*s) {
    if(*s == '{') {
      /* If this is first character, eliminate surrounding braces */
      if(e == buff) {
        chr=1;
        s++;
      } else {
        chr=0;
        *e++=*s++;
      }

      depth=1;
      while(*s && depth) {
        if(*s == '{')
          depth++;
        else if(*s == '}')
          depth--;
        *e++=*s++;
      }

      /* remove closing brace */
      if(chr && *(s-1) == '}')
        end=--e;
      continue;
    }

    if(*s == '[' && type == 1) {  /* Skip over nested function calls */
      *e++=*s++;
      depth=1;
      while(*s && depth) {
        if(*s == '[')
          depth++;
        else if(*s == ']')
          depth--;
        *e++=*s++;
      }
      continue;
    }

    if(*s == '(') {  /* Function Header */
      if(!func) {
        *e='\0';
        *str=s+1;

        if(function(str, buff, privs, doer)) {
          e=end=temp;  /* all future writes will be ignored */
          s=*str;
          func=2;
          continue;
        }
        func=1;
      }

      /* copy text in parentheses literally */
      *e++=*s++;
      depth=1;
      while(*s && depth) {
        if(*s == '(')
          depth++;
        else if(*s == ')')
          depth--;
        *e++=*s++;
      }
      continue;
    }

    /* Check end of function scope */
    if(type == 1 && (*s == ']' || *s == ')' || *s == ','))
      break;

    /* Check delimiter in command-parsing */
    if(type != 1 && *s == type) {
      s++;
      break;
    }

    if(*s)
      *e++=*s++;
  }

  /* remove following spaces (only if there wasn't an ending brace) */
  if(func != 2)
    while(e > end && *(e-1) == ' ')
      e--;

  *e='\0';
  *str=s;
}
